#include "stdafx.h"
#include <windows.h>
#include <stdint.h>
#include <stdio.h>
#include <xinput.h>
#include <dsound.h>
#include <math.h>
#include <malloc.h>
#include <string>
#include <iostream>
#include <sstream>

#define internal static
#define local_persist static
#define global_variable static

/********************** MACROS *************************/
#define ArrayCount(Value) (sizeof(Value) / sizeof(Value[0])) 

#define X_INPUT_GET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_STATE *pState)
typedef X_INPUT_GET_STATE(x_input_get_state);
X_INPUT_GET_STATE(XInputGetStateStub)
{
	return ERROR_DEVICE_NOT_CONNECTED;
}
global_variable x_input_get_state *XInputGetState_ = XInputGetStateStub;
#define XInputGetState XInputGetState_

#define X_INPUT_SET_STATE(name) DWORD WINAPI name(DWORD dwUserIndex, XINPUT_VIBRATION *pVibration)
typedef X_INPUT_SET_STATE(x_input_set_state);
X_INPUT_SET_STATE(XInputSetStateStub)
{
	return ERROR_DEVICE_NOT_CONNECTED;
}
global_variable x_input_set_state *XInputSetState_ = XInputSetStateStub;
#define XInputSetState XInputSetState_

/********************** CONTROLLER STRUCTS *************************/
struct game_button_state
{
	int HalfTransitionCount;
	bool EndedDown;
};

struct game_controller_input
{
	bool IsConnected;

	float StickAverageLX;
	float StickAverageLY;

	float StickAverageRX;
	float StickAverageRY;

	BYTE RightTrigger;
	BYTE LeftTrigger;

	union
	{
		game_button_state Buttons[12];

		struct
		{
			game_button_state Start;
			game_button_state Back;

			game_button_state MoveUp;
			game_button_state MoveDown;
			game_button_state MoveLeft;
			game_button_state MoveRight;

			game_button_state ActionUp;
			game_button_state ActionDown;
			game_button_state ActionLeft;
			game_button_state ActionRight;

			game_button_state LeftShoulder;
			game_button_state RightShoulder;
		};
	};
};

struct game_input
{
	game_controller_input Controllers[4];
};

inline game_controller_input *GetController(game_input *Input, int unsigned ControllerIndex)
{
	game_controller_input *Result = &Input->Controllers[ControllerIndex];
	return Result;
}

/********************** SERIAL COMMS *************************/
#using <System.dll>
using namespace System::IO::Ports;

public ref class PortChat
{
private:
	static SerialPort^ _serialPort;

public:
	static void SerialInit()
	{		
		// Create a new SerialPort object with default settings.
		_serialPort = gcnew SerialPort();

		// Allow the user to set the appropriate properties.
		_serialPort->PortName = "COM4";
		_serialPort->BaudRate = 9600;
		_serialPort->Parity = (Parity)System::Enum::Parse(Parity::typeid, "None");
		_serialPort->DataBits = 8;
		_serialPort->StopBits = (StopBits)System::Enum::Parse(StopBits::typeid, "One");
		_serialPort->Handshake = (Handshake)System::Enum::Parse(Handshake::typeid, "RequestToSend");

		// Set the read/write timeouts
		_serialPort->ReadTimeout = 500;
		_serialPort->WriteTimeout = 50000;

		_serialPort->Open();
	}

	static void SerialWrite(System::String ^*SerialStringToSend)
	{
		_serialPort->WriteLine(*SerialStringToSend);
	}

	static void SerialClose()
	{
		_serialPort->Close();
	}
};

/********************** NUMBER FORMATTING *************************/
internal std::string ZeroPadNumber(uint32_t Number)
{
	std::stringstream ss;

	// Convert Number to string (Stringstream expects 32-bit value)
	ss << Number;
	std::string Result;
	ss >> Result;

	// Prepend "0" to the String if it is less than 3 digits
	int str_length = Result.length();
	for (int i = 0; i < 3 - str_length; i++)
	{
		Result = "0" + Result;
	}

	return Result;
}

/********************** LOAD XINPUT LIBRARY *************************/
internal void Win32LoadXInput(void)
{
	HMODULE XInputLibrary = LoadLibrary(L"xinput1_4.dll");

	if (!XInputLibrary)
	{
		HMODULE XInputLibrary = LoadLibrary(L"xinput9_1_0.dll");
	}

	if (!XInputLibrary)
	{
		HMODULE XInputLibrary = LoadLibrary(L"xinput1_3.dll");
	}

	if (XInputLibrary)
	{
		XInputGetState = (x_input_get_state *)GetProcAddress(XInputLibrary, "XInputGetState");
		XInputSetState = (x_input_set_state *)GetProcAddress(XInputLibrary, "XInputSetState");
	}
}

/********************** DIGITAL BUTTON PROCESSING *************************/
internal void Win32ProcessXInputDigitalButton(DWORD XInputButtonState, game_button_state *OldState, DWORD ButtonBit, game_button_state *NewState)
{
	NewState->EndedDown = (XInputButtonState & ButtonBit) == ButtonBit;
	NewState->HalfTransitionCount = (OldState->EndedDown != NewState->EndedDown) ? 1 : 0;
}

/********************** ANALOG STICK PROCESSING *************************/
internal float Win32ProcessXInputStickValue(SHORT Value, SHORT DeadZoneThreshold)
{
	float Result = 0;

	if (Value < -DeadZoneThreshold)
	{
		Result = (float)((Value + DeadZoneThreshold) / (32768.0f - DeadZoneThreshold));
	}
	else if (Value > DeadZoneThreshold)
	{
		Result = (float)((Value - DeadZoneThreshold) / (32767.0f - DeadZoneThreshold));
	}

	return Result;
}

/********************** VARIABLE UPDATE *************************/
internal void ControllerUpdateAndOutput(game_input *Input)
{
	bool Brake = false;
	bool Left = false;
	bool Right = false;
	BYTE ForwardSpeed = 0;
	BYTE ReverseSpeed = 0;

	game_controller_input *Controller = GetController(Input, 0);
		
	if (Controller->ActionDown.EndedDown)
	{
		Brake = true;
	}
	if (Controller->MoveLeft.EndedDown)
	{
		Left = true;
	}
	if (Controller->MoveRight.EndedDown)
	{
		Right = true;
	}

	ForwardSpeed = Controller->RightTrigger;
	ReverseSpeed = Controller->LeftTrigger;

	//String to send over Serial: 3-digits for ForwardSpeed, 3-digits for ReverseSpeed, 1-digit for Left, 1-digit for Right, 1-digit for brake
	//Forward Reverse Left Right Brake
	//  xxx     xxx    x     x     x

	//Formatting numbers with zero-padding
	std::string ForwardSpeedString = ZeroPadNumber((uint32_t)ForwardSpeed);
	std::string ReverseSpeedString = ZeroPadNumber((uint32_t)ReverseSpeed);

	//Converting std::string to System::String^ since Serial Write expects this type
	System::String^ ForwardString = gcnew System::String(ForwardSpeedString.c_str());
	System::String^ ReverseString = gcnew System::String(ReverseSpeedString.c_str());

	//Convert Left, Right and Brake to System::String^
	System::String^ LeftString = ((int)(Left)).ToString();
	System::String^ RightString = ((int)(Right)).ToString();
	System::String^ BrakeString = ((int)(Brake)).ToString();

	//Concatenate String
	System::String^ SerialStringToSend = "1111x" + ForwardString + "x" + ReverseString + "x" + LeftString + "x" + RightString + "x" + BrakeString + "x";

	//Console Debug
	System::Console::Write("Forward: " + ForwardString + ", " + "Backward: " + ReverseString + ", " + "Left: " + LeftString + ", " + "Right: " + RightString + ", " + "Brake: " + BrakeString + "\r");

	//Send String out over Serial Comms
	PortChat::SerialWrite(&SerialStringToSend);
}

/********************** MAIN LOOP *************************/
int _tmain(int argc, _TCHAR* argv[])
{
	Win32LoadXInput();
	PortChat::SerialInit();

	game_input Input[2] = {};
	game_input *NewInput = &Input[0];
	game_input *OldInput = &Input[1];

	bool Running = true;

	while(Running)
	{
		DWORD MaxControllerCount = XUSER_MAX_COUNT;

		//TODO(devon): should we poll this more frequently
		for(DWORD ControllerIndex = 0; ControllerIndex < MaxControllerCount; ++ControllerIndex)
		{
			DWORD OurControllerIndex = ControllerIndex;
			game_controller_input *OldController = GetController(OldInput, OurControllerIndex);
			game_controller_input *NewController = GetController(NewInput, OurControllerIndex);

			XINPUT_STATE ControllerState;
			if(XInputGetState(ControllerIndex, &ControllerState) == ERROR_SUCCESS)
			{
				NewController->IsConnected = true;

				//NOTE(devon): controller is plugged in
				XINPUT_GAMEPAD *Pad = &ControllerState.Gamepad;

				//Analog Sticks
				NewController->StickAverageLX = Win32ProcessXInputStickValue(Pad->sThumbLX, XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE);
				NewController->StickAverageLY = Win32ProcessXInputStickValue(Pad->sThumbLY, XINPUT_GAMEPAD_LEFT_THUMB_DEADZONE);
				NewController->StickAverageRX = Win32ProcessXInputStickValue(Pad->sThumbRX, XINPUT_GAMEPAD_RIGHT_THUMB_DEADZONE);
				NewController->StickAverageRY = Win32ProcessXInputStickValue(Pad->sThumbRY, XINPUT_GAMEPAD_RIGHT_THUMB_DEADZONE);

				NewController->RightTrigger = Pad->bRightTrigger;
				NewController->LeftTrigger = Pad->bLeftTrigger;

				//DPAD Control
				if(Pad->wButtons & XINPUT_GAMEPAD_DPAD_UP)
				{
					NewController->StickAverageLY = 1.0f;
				}
				if(Pad->wButtons & XINPUT_GAMEPAD_DPAD_DOWN)
				{
					NewController->StickAverageLY = -1.0f;
				}
				if(Pad->wButtons & XINPUT_GAMEPAD_DPAD_LEFT)
				{
					NewController->StickAverageLX = -1.0f;
				}
				if(Pad->wButtons & XINPUT_GAMEPAD_DPAD_RIGHT)
				{
					NewController->StickAverageLX = 1.0f;
				}

				//Analog Sticks to Digital Buttons
				float Threshold = 0.5f;

				Win32ProcessXInputDigitalButton((NewController->StickAverageLX < -Threshold) ? 1 : 0, &OldController->MoveLeft, 1, &NewController->MoveLeft);
				Win32ProcessXInputDigitalButton((NewController->StickAverageLX > Threshold) ? 1 : 0, &OldController->MoveRight, 1, &NewController->MoveRight);
				Win32ProcessXInputDigitalButton((NewController->StickAverageLY < -Threshold) ? 1 : 0, &OldController->MoveDown, 1, &NewController->MoveDown);
				Win32ProcessXInputDigitalButton((NewController->StickAverageLY > Threshold) ? 1 : 0, &OldController->MoveUp, 1, &NewController->MoveUp);

				Win32ProcessXInputDigitalButton((NewController->StickAverageRX < -Threshold) ? 1 : 0, &OldController->ActionLeft, 1, &NewController->ActionLeft);
				Win32ProcessXInputDigitalButton((NewController->StickAverageRX > Threshold) ? 1 : 0, &OldController->ActionRight, 1, &NewController->ActionRight);
				Win32ProcessXInputDigitalButton((NewController->StickAverageRY < -Threshold) ? 1 : 0, &OldController->ActionDown, 1, &NewController->ActionDown);
				Win32ProcessXInputDigitalButton((NewController->StickAverageRY > Threshold) ? 1 : 0, &OldController->ActionUp, 1, &NewController->ActionUp);

				//Digital Buttons
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->Start, XINPUT_GAMEPAD_START, &NewController->Start);
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->Back, XINPUT_GAMEPAD_BACK, &NewController->Back);
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->LeftShoulder, XINPUT_GAMEPAD_LEFT_SHOULDER, &NewController->LeftShoulder);
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->RightShoulder, XINPUT_GAMEPAD_RIGHT_SHOULDER, &NewController->RightShoulder);
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->ActionDown, XINPUT_GAMEPAD_A, &NewController->ActionDown);
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->ActionRight, XINPUT_GAMEPAD_B, &NewController->ActionRight);
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->ActionLeft, XINPUT_GAMEPAD_X, &NewController->ActionLeft);
				Win32ProcessXInputDigitalButton(Pad->wButtons, &OldController->ActionUp, XINPUT_GAMEPAD_Y, &NewController->ActionUp);
			}
			else
			{
				//NOTE(devon): controller is not available
				NewController->IsConnected = false;
			}

			if (NewController->Start.EndedDown)
			{
				Running = false;
			}
		}

		ControllerUpdateAndOutput(&Input[0]);

		//Variable Swap to reset for next time through loop
		game_input *Temp = NewInput;
		NewInput = OldInput;
		OldInput = Temp;
	}

	PortChat::SerialClose();

	return 0;
}
